/* Implementation of forward secure signatures with untrusted update.
 * Copyright 2006 Emily Shen, licensed under the GPLv2 as described in COPYING
 */

#include "fssig.h"
#include <math.h>
#include <stdio.h>

int fs_verify(unsigned char *sig, unsigned char *msg, 
	      unsigned int msg_len_bits, int time, fs_public_key_t pk)
{
  int result;
  int len1, len2;
  pairing_ptr pairing = pk->param->pairing;
  unsigned int i;
  int *time_bits, *msg_bits;
  element_t lhs, rhs, t0, t1, prod_ui, prod_vi;

  element_t sigma1, sigma2, sigma3;
  element_init_G2(sigma1, pairing);
  element_init_G1(sigma2, pairing);
  element_init_G1(sigma3, pairing);
  len1 = element_from_bytes(sigma1, sig);
  len2 = element_from_bytes(sigma2, sig + len1);
  element_from_bytes(sigma3, sig + len1 + len2);

  time_bits = malloc(pk->t * sizeof(int));
  for (i = 0; i < (unsigned int) pk->t; i++) {
    time_bits[pk->t - i - 1] = ((time & (1 << i)) != 0);
  }

  msg_bits = malloc(msg_len_bits * sizeof(int));
  for (i = 0; i < msg_len_bits; i++) {
    msg_bits[i] = ((msg[i / 8] & (1 << (i % 8))) != 0);
  }
  
  element_init_GT(lhs, pairing);
  pairing_apply(lhs, pk->g1, sigma1, pairing);
  
  element_init_G2(prod_ui, pairing);
  element_set(prod_ui, pk->u_prime);
  for (i = 0; i < (unsigned int) pk->t; i++) {
    if (time_bits[i]) {
      element_mul(prod_ui, prod_ui, pk->U[i]);
    }
  }

  element_init_G2(prod_vi, pairing);
  element_set(prod_vi, pk->v_prime);
  for (i = 0; i < msg_len_bits; i++) {
    if (msg_bits[i]) {
      element_mul(prod_vi, prod_vi, pk->V[i]);
    }
  }

  element_init_GT(rhs, pairing);
  element_set(rhs, pk->e_g1_g2_v);
  element_init_GT(t0, pairing);
  element_init_GT(t1, pairing);
  pairing_apply(t0, sigma2, prod_ui, pairing);
  pairing_apply(t1, sigma3, prod_vi, pairing);
  element_mul(rhs, rhs, t0);
  element_mul(rhs, rhs, t1);
  
  result = !element_cmp(lhs, rhs);

  element_clear(lhs);
  element_clear(rhs);
  element_clear(t0);
  element_clear(t1);
  element_clear(prod_ui);
  element_clear(prod_vi);
  free(time_bits);
  free(msg_bits);

  return result;
}

sk_node_ptr get_curr_leaf(fs_private_key_t sk)
{
  int i;
  int *time_bits = malloc(sk->t * sizeof(int));
  sk_node_ptr curr = sk->root;
  for (i = 0; i < sk->t; i++) {
    time_bits[sk->t - i - 1] = ((sk->time & (1 << i)) != 0);
  }
  for (i = 0; i < sk->t; i++) {
    if (time_bits[i] == 0) {
      curr = curr->left;
    }
    else {
      curr = curr->right;
    }
  }
  free(time_bits);
  return curr;
}

void get_sk(fs_private_key_t sk, element_t s1, element_t s2)
{
  sk_node_ptr curr = get_curr_leaf(sk);
  element_set(s1, curr->sk_vector[0]);
  element_set(s2, curr->sk_vector[1]);
}

void fs_sign(unsigned char *sig, unsigned char *msg, unsigned int msg_len_bits,
	     fs_public_key_t pk, fs_private_key_t sk, fs_dec_key_t dk)
{ 
  pairing_ptr pairing = pk->param->pairing;
  element_t r, r2, s1, s2, sigma1, sigma2, sigma3, prod_vi, prod_ui;
  int *msg_bits;
  int len1, len2;
  unsigned int i;
  int *time_bits;
  
  element_init_G2(s1, pairing);
  element_init_G1(s2, pairing);

  get_sk(sk, s1, s2);

  msg_bits = malloc(msg_len_bits * sizeof(int));
  for (i = 0; i < msg_len_bits; i++) {
    msg_bits[i] = ((msg[i / 8] & (1 << (i % 8))) != 0);
  }

  element_init_G2(prod_vi, pairing);
  element_set(prod_vi, pk->v_prime);
  for (i = 0; i < msg_len_bits; i++) {
    if (msg_bits[i]) {
      element_mul(prod_vi, prod_vi, pk->V[i]);
    }
  }

  time_bits = malloc(pk->t * sizeof(int));
  for (i = 0; i < (unsigned int) pk->t; i++) {
    time_bits[pk->t - i - 1] = ((sk->time & (1 << i)) != 0);
  }

  element_init_G2(prod_ui, pairing);
  element_set(prod_ui, pk->u_prime);
  for (i = 0; i < (unsigned int) pk->t; i++) {
    if (time_bits[i]) {
      element_mul(prod_ui, prod_ui, pk->U[i]);
    }
  }

  element_init_Zr(r, pairing);
  element_random(r);

  element_pow_zn(prod_vi, prod_vi, r);

  element_init_Zr(r2, pairing);
  element_random(r2);

  element_pow_zn(prod_ui, prod_ui, r2);

  element_init_G2(sigma1, pairing);
  element_init_G1(sigma2, pairing);
  element_init_G1(sigma3, pairing);

  element_mul(sigma1, s1, prod_vi);
  element_mul(sigma1, sigma1, prod_ui);
  element_mul(sigma1, sigma1, dk->dk);
  element_pow_zn(sigma2, pk->g1, r2);
  element_mul(sigma2, sigma2, s2);
  element_pow_zn(sigma3, pk->g1, r);

  len1 = element_to_bytes(sig, sigma1);
  len2 = element_to_bytes(sig + len1, sigma2);
  element_to_bytes(sig + len1 + len2, sigma3);

  element_clear(r);
  element_clear(r2);
  element_clear(s1);
  element_clear(s2);
  element_clear(sigma1);
  element_clear(sigma2);
  element_clear(sigma3);
  element_clear(prod_vi);
  element_clear(prod_ui);

  free(msg_bits);
  free(time_bits);
}

void fs_update(fs_private_key_t sk, fs_public_key_t pk)
{
  int i;
  int *time_bits;
  element_t *left_vector, *right_vector;
  pairing_ptr pairing = pk->param->pairing;
  sk_node_ptr curr_leaf = get_curr_leaf(sk);
  sk_node_ptr curr = curr_leaf;
  free(curr_leaf->sk_vector);
  curr_leaf->sk_vector = NULL;
  element_t s, prod_ui, prod_ui_s;
  element_init_Zr(s, pairing);
  
  sk->time++;
  if (sk->time >= sk->T) {
    sk->root = NULL;
    return;
  }
  
  while (curr == curr->parent->right) {
    curr = curr->parent;
  }
  curr = curr->parent->right;
  time_bits = malloc(sk->t * sizeof(int));
  for (i = 0; i < sk->t; i++) {
    time_bits[sk->t - i - 1] = ((sk->time & (1 << i)) != 0);
  }

  element_init_G2(prod_ui, pairing);
  element_set(prod_ui, pk->u_prime);
  for (i = 0; i < sk->t; i++) {
    if (time_bits[i]) {
      element_mul(prod_ui, prod_ui, pk->U[i]);
    }
  }
  element_init_G2(prod_ui_s, pairing);

  while (curr->depth < sk->t) {
    curr->left = malloc(sizeof(struct sk_node_s));
    curr->right = malloc(sizeof(struct sk_node_s));
    curr->left->left = curr->left->right = curr->right->left
      = curr->right->right = NULL;
    curr->left->parent = curr->right->parent = curr;
    curr->left->depth = curr->right->depth = curr->depth + 1;
    curr->left->sk_vector = malloc((2 + sk->t - curr->left->depth) * 
				   sizeof(element_t));
    curr->right->sk_vector = malloc((2 + sk->t - curr->right->depth) * 
				    sizeof(element_t));
    left_vector = curr->left->sk_vector;
    right_vector = curr->right->sk_vector;
    element_random(s);
    element_init_G2(left_vector[0], pairing);
    element_init_G2(right_vector[0], pairing);

    element_pow_zn(prod_ui_s, prod_ui, s);
    element_mul(left_vector[0], prod_ui_s, curr->sk_vector[0]);
    element_pow_zn(right_vector[0], pk->U[curr->depth], s);
    element_mul(right_vector[0], right_vector[0], curr->sk_vector[2]);
    element_mul(right_vector[0], right_vector[0], left_vector[0]);
    element_init_G1(left_vector[1], pairing);
    element_init_G1(right_vector[1], pairing);
    element_pow_zn(left_vector[1], pk->g1, s);
    element_mul(left_vector[1], left_vector[1], curr->sk_vector[1]);
    element_pow_zn(right_vector[1], pk->g1, s);
    element_mul(right_vector[1], right_vector[1], curr->sk_vector[1]);
    
    for (i = 0; i < sk->t - curr->left->depth; i++) {
      element_init_G2(left_vector[i+2], pairing);
      element_init_G2(right_vector[i+2], pairing);
      element_pow_zn(left_vector[i+2], pk->U[i+curr->left->depth], s);
      element_mul(left_vector[i+2], left_vector[i+2], curr->sk_vector[i+3]);
      element_pow_zn(right_vector[i+2], pk->U[i+curr->left->depth], s);
      element_mul(right_vector[i+2], right_vector[i+2], curr->sk_vector[i+3]);
    }
    free(curr->sk_vector);
    curr->sk_vector = NULL;
    curr = curr->left;
  }
  free(time_bits);
  element_clear(s);
  element_clear(prod_ui);
  element_clear(prod_ui_s);
}

void fs_gen_sys_param(fs_sys_param_t param, pairing_t pairing)
{
  param->pairing = pairing;
  param->sig_length = 2 * pairing->G1->fixed_length_in_bytes 
      + pairing->G2->fixed_length_in_bytes;
}

void fs_gen(fs_public_key_t pk, fs_private_key_t sk, fs_dec_key_t dk, fs_sys_param_t param, 
	    unsigned int msg_len_bits, unsigned int num_intervals)
{
  element_t v, w, alpha, neg_w, r, s0, e_g1_g2;
  unsigned int i, j;
  sk_node_ptr curr;
  pk->param = sk->param = param;
  pk->T = sk->T = num_intervals;
  pk->t = sk->t = log(num_intervals) / log(2.0) + 0.5;
  sk->time = 0;
  element_init_G1(pk->g1, param->pairing);
  element_random(pk->g1);
  element_init_G2(pk->g2, param->pairing);
  element_random(pk->g2);
  element_init_Zr(v, param->pairing);
  element_init_Zr(w, param->pairing);
  element_random(v);
  element_random(w);
  element_init_Zr(neg_w, param->pairing);
  element_neg(neg_w, w);
  element_init_Zr(alpha, param->pairing);
  element_add(alpha, v, w);
  element_init_G2(dk->dk, param->pairing);
  element_pow_zn(dk->dk, pk->g2, neg_w);
  element_init_G2(pk->u_prime, param->pairing);
  element_random(pk->u_prime);
  element_init_G2(pk->v_prime, param->pairing);
  element_random(pk->v_prime);
  pk->U = malloc(pk->t * sizeof(element_t));
  for (i = 0; i < (unsigned int) pk->t; i++) {
    element_init_G2(pk->U[i], param->pairing);
    element_random(pk->U[i]);
  }
  pk->V = malloc(msg_len_bits * sizeof(element_t));
  for (i = 0; i < msg_len_bits; i++) {
    element_init_G2(pk->V[i], param->pairing);
    element_random(pk->V[i]);
  }

  element_init_GT(e_g1_g2, param->pairing);
  element_init_GT(pk->e_g1_g2_v, param->pairing);
  pairing_apply(e_g1_g2, pk->g1, pk->g2, param->pairing);
  element_pow_zn(pk->e_g1_g2_v, e_g1_g2, v);
  element_init_GT(pk->e_g1_g2_w, param->pairing);
  element_pow_zn(pk->e_g1_g2_w, e_g1_g2, w);

  sk->root = malloc(sizeof(struct sk_node_s));
  sk->root->sk_vector = NULL;
  sk->root->depth = 0;
  sk->root->parent = NULL;
  
  curr = sk->root;
  element_init_Zr(r, param->pairing);
  element_init_G2(s0, param->pairing);
  for (i = 0; i < (unsigned int) pk->t; i++) {
    curr->left = malloc(sizeof(struct sk_node_s));
    curr->right = malloc(sizeof(struct sk_node_s));
    curr->left->sk_vector = NULL;
    curr->left->left = curr->left->right = curr->right->left 
      = curr->right->right = NULL;
    curr->left->parent = curr->right->parent = curr;
    curr->left->depth = curr->right->depth = curr->depth + 1;
    
    element_random(r);
    curr->right->sk_vector = malloc((1 + pk->t - i) * sizeof(element_t));
    element_init_G2(curr->right->sk_vector[0], param->pairing);
    element_init_G1(curr->right->sk_vector[1], param->pairing);
    for (j = 2; j < (1 + pk->t - i); j++) {
      element_init_G2(curr->right->sk_vector[j], param->pairing);
      element_pow_zn(curr->right->sk_vector[j], pk->U[i+j-1], r);
    }
    element_mul(s0, pk->u_prime, pk->U[i]);
    element_pow_zn(s0, s0, r);
    element_pow_zn(curr->right->sk_vector[0], pk->g2, alpha);
    element_mul(curr->right->sk_vector[0], curr->right->sk_vector[0], s0);
    element_pow_zn(curr->right->sk_vector[1], pk->g1, r);
    
    curr = curr->left;
  }

  curr->sk_vector = malloc(2 * sizeof(element_t));
  element_init_G2(curr->sk_vector[0], param->pairing);
  element_init_G1(curr->sk_vector[1], param->pairing);
  element_pow_zn(curr->sk_vector[0], pk->g2, alpha);
  element_pow_zn(s0, pk->u_prime, r);
  element_mul(curr->sk_vector[0], curr->sk_vector[0], s0);
  element_pow_zn(curr->sk_vector[1], pk->g1, r);

  element_clear(v);
  element_clear(w);
  element_clear(alpha);
  element_clear(neg_w);
  element_clear(r);
  element_clear(s0);
  element_clear(e_g1_g2);
}

int fs_check_key(fs_private_key_t sk, fs_public_key_t pk)
{
  element_t s1, s2, prod_ui, lhs, rhs;
  pairing_ptr pairing = pk->param->pairing;
  int *time_bits;
  unsigned int i;
  int result;

  if (sk->time >= sk->T) {
    return 0;
  }
  element_init_G2(s1, pairing);
  element_init_G1(s2, pairing);
  get_sk(sk, s1, s2);

  time_bits = malloc(pk->t * sizeof(int));
  for (i = 0; i < (unsigned int) pk->t; i++) {
    time_bits[pk->t - i - 1] = ((sk->time & (1 << i)) != 0);
  }

  element_init_G2(prod_ui, pairing);
  element_set(prod_ui, pk->u_prime);
  for (i = 0; i < (unsigned int) pk->t; i++) {
    if (time_bits[i]) {
      element_mul(prod_ui, prod_ui, pk->U[i]);
    }
  }
  
  element_init_GT(lhs, pairing);
  element_init_GT(rhs, pairing);
  pairing_apply(lhs, pk->g1, s1, pairing);
  pairing_apply(rhs, s2, prod_ui, pairing);
  element_mul(rhs, rhs, pk->e_g1_g2_v);
  element_mul(rhs, rhs, pk->e_g1_g2_w);

  result = !element_cmp(lhs, rhs);
  element_clear(s1);
  element_clear(s2);
  element_clear(prod_ui);
  element_clear(lhs);
  element_clear(rhs);
  return result;
}
